// The game of Asteroids.

using Cairo;
using Gdk;
using GLib;
using Gtk;

using System;
using System.Collections.Generic;
using static System.Console;
using static System.Math;

using Color = Gdk.Color;
using Key = Gdk.Key;
using Timeout = GLib.Timeout;
using Window = Gtk.Window;

struct Point {
  public readonly double x, y;
  public Point(double x, double y) { this.x = x; this.y = y; }

  public static Point Origin {
    get => new Point();
  }

  public static Point operator + (Point p, Vec v) =>
    new Point(p.x + v.dx, p.y + v.dy);
    
  public static Vec operator - (Point p, Point q) =>
    new Vec(p.x - q.x, p.y - q.y);
    
  public Point rotate(double r) =>    // around the origin
    new Point(x * Cos(r) - y * Sin(r), x * Sin(r) + y * Cos(r));
  
  // Convert this point to a Cairo PointD for convenience in calling Cairo methods.
  public PointD pointd { get => new PointD(x, y); }

  // Wrap each coordinate of this Point into the range [0, d).
  public static Point operator % (Point p, double d) =>
    new Point(((p.x % d) + d) % d, ((p.y % d) + d) % d);
}

struct Vec {
  public readonly double dx, dy;
  public Vec(double dx, double dy) { this.dx = dx; this.dy = dy; }
  
  public static Vec operator + (Vec v, Vec w) =>
    new Vec(v.dx + w.dx, v.dy + w.dy);
    
  public static Vec operator * (double d, Vec v) =>
    new Vec(d * v.dx, d * v.dy);
    
  public double length => Sqrt(dx * dx + dy + dy);
  
  // Return the z-component of the cross product of v and w.
  public static double cross(Vec v, Vec w) =>
    v.dx * w.dy - v.dy * w.dx;

  // Return a unit vector pointing in the given direction.
  public static Vec direction(double theta) =>
    new Vec(Cos(theta), Sin(theta));
}

struct Segment {
  public readonly Point p, q;
  public Segment(Point p, Point q) { this.p = p; this.q = q; }
  
  // Return true if this segment straddles the line containing s.
  bool straddle(Segment s) =>
    Vec.cross(p - s.p, s.q - s.p) * Vec.cross(q - s.p, s.q - s.p) < 0;
    
  // Check if this segment intersects s.
  // (See Introduction to Algorithms, 33.1 "Line-segment properties".)
  public bool intersect(Segment s) =>
    s.straddle(this) && this.straddle(s);
}

struct Polygon {
  public readonly Point[] a;
  public Polygon(Point[] a) { this.a = a; }
  
  Segment seg(int i) => new Segment(a[i], a[(i + 1) % a.Length]);
  
  public Polygon transform(Vec translate, double rot) {
    Point[] b = new Point[a.Length];
    for (int i = 0 ; i < a.Length ; ++i)
      b[i] = a[i].rotate(rot) + translate;
    return new Polygon(b);
  }
  
  // Return true if polygons p and q intersect.  We look for a pair of intersecting line
  // segments (and don't check whether one polygon might be inside the other).
  public static bool intersect(Polygon p, Polygon q) {
    for (int i = 0 ; i < p.a.Length ; ++i) {
      Segment s = p.seg(i);
      for (int j = 0 ; j < q.a.Length ; ++j)
        if (s.intersect(q.seg(j)))
          return true;
    }
    return false;
  }
}

class Util {
  public static double radians(double deg) => deg / 360.0 * 2 * PI;
}

abstract class Thing {
  public Point pos;     // position
  public Vec vel;       // velocity
  public double rot;    // rotation (clockwise)
    
  protected abstract Polygon geometry();
  
  // Return this Thing's polygon, transformed to position (pos) and rotated by (rot).
  public Polygon polygon() => geometry().transform(pos - Point.Origin, rot);
}

class Ship : Thing {
  public bool thrust;
  public int turning;   // 1 = clockwise, 0 = not turning, -1 = counterclockwise
  
  public Ship(double x, double y) {
    pos = new Point(x, y);
  }
  
  static Polygon Shape = new Polygon(new Point[] {
    new Point(40, 0), new Point(-10, 10), new Point(-10, -10)
  });
  
  protected override Polygon geometry() => Shape;
  
  public void move() {
    rot += turning * Util.radians(15);
    
    if (thrust)
      vel = vel + 1.0 * Vec.direction(rot);
    else vel = 0.99 * vel;    // friction
    
    pos = (pos + vel) % Game.Size;
  }
}

class Bullet : Thing {
  public Bullet(Point pos, Vec vel) {
    this.pos = pos; this.vel = vel;
  }
  
  static Polygon Shape = new Polygon(new Point[] {
    new Point(4, 0), new Point(0, 4), new Point(-4, 0), new Point(0, -4)
  });  
  
  protected override Polygon geometry() => Shape;

  // Move this bullet, returning true if has moved off the screen.
  public bool move() {
    pos += vel;
    rot += 0.2;
    return (pos.x < 0 || pos.x >= Game.Size ||
            pos.y < 0 || pos.y >= Game.Size);
  }
}

class Asteroid : Thing {
  public int size;  // 2 = large, 1 = medium, 0 = small
  Polygon shape;
  
  static int[] Radius = { 10, 20, 40 };
  
  public Asteroid(int size, Point pos) {
    this.size = size;
    this.pos = pos;
    
    int radius = Radius[size];
    shape = new Polygon(new Point[] {
      new Point(radius, 0), new Point(0, radius),
      new Point(- radius, 0), new Point(0, - radius)
    });  
    
    Random random = Game.random;
    rot = random.NextDouble() * 2 * PI;
    
    double theta = random.NextDouble() * 2 * PI;
    double speed = 1.0 + random.NextDouble() * 4.0;
    vel = speed * Vec.direction(theta);
  }
  
  protected override Polygon geometry() => shape;

  public void move() {
    pos = (pos + vel) % Game.Size;
  }
}

class Game {
  public const int Size = 600;
  
  public Ship ship = new Ship(Size / 2, Size / 2);
  public List<Bullet> bullets = new List<Bullet>();
  public List<Asteroid> asteroids = new List<Asteroid>();
  public bool dead;
  
  public static Random random = new Random();
  
  public Game() {
    for (int i = 0 ; i < 5 ; ++i) {
      Point pos;
      
      // Choose an asteroid position that's not too close to the ship.
      // (This does not consider wraparound, but the ship starts in the center
      // so that is OK.)
      do {
        pos = new Point(random.Next(Size), random.Next(Size));
      } while ((pos - ship.pos).length < 200);
      
      asteroids.Add(new Asteroid(2 /* large */, pos));
    }
  }
  
  // Check if any bullet has hit asteroid i.  Return true if so.
  bool hitAsteroid(int i) {
    Asteroid a = asteroids[i];
    Polygon ap = a.polygon();
    
    for (int j = 0 ; j < bullets.Count ; ++j)
      if (Polygon.intersect(bullets[j].polygon(), ap)) {
        bullets.RemoveAt(j);
        asteroids.RemoveAt(i);
        if (a.size > 0)  // the asteroid was not small
          for (int k = 0 ; k < 2 ; ++k)  // so create two fragments
            asteroids.Add(new Asteroid(a.size - 1, a.pos));
        return true;
      }
      
    return false;
  }
  
  // Move the game forward by one tick.
  public void tick() {
    ship.move();
    
    int i = 0;
    while (i < bullets.Count)
      if (bullets[i].move())
        bullets.RemoveAt(i);
      else ++i;
    
    // Move each asteroid, and check if it has hit the player or a bullet.
    Polygon sp = ship.polygon();
    i = 0;
    while (i < asteroids.Count) {
      Asteroid a = asteroids[i];
      a.move();
      if (Polygon.intersect(sp, a.polygon()))
        dead = true;
        
      if (!hitAsteroid(i))
        ++i;
    }
  }
  
  public void fire() {
    if (bullets.Count < 4)
      bullets.Add(new Bullet(ship.pos + 30.0 * Vec.direction(ship.rot),
                             ship.vel + 10.0 * Vec.direction(ship.rot)));
  }
}

static class GtkUtil {
  public static Color parse(string name) {
    Color c = new Color(0, 0, 0);
    if (!Color.Parse(name, ref c))
      throw new Exception("bad color");
    return c;
  }
  
  public static void setColor(this Context c, Color color) {
    CairoHelper.SetSourceColor(c, color);
  }
}

class View : DrawingArea {
  Game game = new Game();
  
  Ship ship { get => game.ship; }
  
  public View() {
    ModifyBg(StateType.Normal, new Gdk.Color(0, 0, 0));  // set background to black
    Timeout.Add(50, timeout);  // ask for a callback every 50 ms, i.e. 20 times per second
  }
  
  public bool timeout() {
    game.tick();
    QueueDraw();
    return true;
  }
  
  // Draw and fill a polygon in the given Cairo context.
  void draw(Context c, Polygon p) {
    c.MoveTo(p.a[0].pointd);
    for (int i = 1 ; i < p.a.Length ; ++i)
      c.LineTo(p.a[i].pointd);
    c.ClosePath();
    c.Fill();
  }
  
  protected override bool OnExposeEvent(EventExpose ev) {
    Context c = CairoHelper.Create(GdkWindow);

    c.setColor(GtkUtil.parse(game.dead ? "red" : "dodger blue"));
    draw(c, ship.polygon());
    
    c.setColor(GtkUtil.parse("red"));
    foreach (Bullet b in game.bullets)
      draw(c, b.polygon());
    
    c.setColor(GtkUtil.parse("gray"));
    foreach (Asteroid a in game.asteroids)
      draw(c, a.polygon());
    
    c.GetTarget().Dispose();
    c.Dispose();
    return false;
  }
  
  public void keyPress(Key key) {
    switch (key) {
      case Key.Left:
        ship.turning = -1; break;
      case Key.Right:
        ship.turning = 1; break;
      case Key.x:
        ship.thrust = true; break;
      case Key.space:
        game.fire(); break;
    }
  }
  
  public void keyRelease(Key key) {
    switch (key) {
      case Key.Left:
        if (ship.turning == -1)
          ship.turning = 0;
        break;
      case Key.Right:
        if (ship.turning == 1)
          ship.turning = 0;
        break;
      case Key.x:
        ship.thrust = false;
      break;
    }
  }
}

public class Frame : Window {
  View view;
  
  public Frame() : base("game") {
   Resize(Game.Size, Game.Size);
   Add(view = new View());
  }
 
  protected override bool OnDeleteEvent(Event ev) {
    Application.Quit();
    return true;
  }
 
  protected override bool OnKeyPressEvent(EventKey ev) {
    view.keyPress(ev.Key);
    return true;
  }
  
  protected override bool OnKeyReleaseEvent(EventKey ev) {
    view.keyRelease(ev.Key);
    return true;
  }
  
 public static void Main() {
   Application.Init();
   new Frame().ShowAll();
   Application.Run();
 }
}
